﻿//======================================================================================================================
// NunchuckVJoy project, version 1.0
// Giliam de Carpentier, Copyright (c) 2019.
// See decarpentier.nl/making-a-wireless-joystick for project details.
//======================================================================================================================
// Source code is licensed under the MIT license. 
// See mit.txt for details.
// Use at your own risk!
//======================================================================================================================
// NOTE: Bluetooth LE API is accessed via the UwpDesktop package, which can be retrieved using NuGet as follows: 
// "Install-Package UwpDesktop -Version 10.0.14393.3"
//======================================================================================================================

using System;
using System.Collections.Generic;
using Windows.Devices.Bluetooth;
using Windows.Devices.Bluetooth.Advertisement;
using Windows.Devices.Bluetooth.GenericAttributeProfile;
using Windows.Storage.Streams;

namespace NunchuckVJoy
{
    // BluetoothConnection starts by continuously scanning for advertisements of paired Bluetooth LE devices, and 
    // once connected, it calls external connect, message and disconnect callbacks at the right moment. 
    // Once disconnect, it will restart the scanning process again.
    public class BluetoothConnection
    {
        // Callback called on every connection attempt
        public event Action<string> OnConnect;

        // Callback called on every disconnect attempt
        public event Action OnDisconnect;

        // Callback called when a message is received
        public event Action<byte[]> OnMessage;

        // Flag indicating if it's currently looking for a new device
        public bool IsInspecting { get; private set; }
        public BluetoothLEDevice Device { get; private set; }

        // Create an instance that will search for a device with a name starting with nameStartsWith and offers a 
        // communication service with the given serviceGuid Guid.
        public BluetoothConnection(string nameStartsWith, Guid serviceGuid)
        {
            _nameStartsWith = nameStartsWith;
            _serviceGuid = serviceGuid;

            _advertisementWatcher.Received += InspectDevice;
        }

        // Stop all activity and dispose all references on destruction
        ~BluetoothConnection()
        {
            Stop();
        }

        // Start looking for a connection
        public void Start()
        {
            if (!IsInspecting)
                _advertisementWatcher.Start();
        }

        // Stop looking for a connection
        public void Stop()
        {
            if (Device != null)
            {
                // Broadcast we're now disconnecting
                var onDisconnect = OnDisconnect;
                if (onDisconnect != null)
                    onDisconnect();

                // Clean up the old connection
                ulong bluetoothAddress = Device.BluetoothAddress;
                var services = Device.GattServices;

                foreach (var service in services)
                    service.Dispose();

                Device.Dispose();
                Device = null;

                lock (_addressesBeingInspected)
                {
                    _addressesBeingInspected.Remove(bluetoothAddress);
                }
            }

            if (IsInspecting)
                _advertisementWatcher.Stop();
        }

        // Callback handling new data received by the bluetooth service
        private void ReceivedData(GattCharacteristic sender, GattValueChangedEventArgs args)
        {
            var buffer = new byte[args.CharacteristicValue.Length];

            DataReader.FromBuffer(args.CharacteristicValue).ReadBytes(buffer);

            var onMessage = OnMessage;
            if (onMessage != null)
                onMessage(buffer);
        }

        // Callback handling disconnects
        private void ConnectionChanged(BluetoothLEDevice device, System.Object sender)
        {
            if (Device != null && device.ConnectionStatus == BluetoothConnectionStatus.Disconnected)
            {
                Stop();

                _advertisementWatcher.Start();
                IsInspecting = true;
            }
        }

        // Possibly parallel asynchronous callback for each device found, carefully managing duplicate calls.
        private async void InspectDevice(BluetoothLEAdvertisementWatcher watcher, BluetoothLEAdvertisementReceivedEventArgs args)
        {
            // When we're already processing this bluetooth address in another async / thread, dont inspect the device here as well. 
            // This is important to prevent exceptions triggered by liftetime issues of (typically the charasteristics of) the BluetoothLEDevice object.
            bool alreadyInspectingDevice = false;
            lock (_addressesBeingInspected)
            {
                if (!_addressesBeingInspected.Contains(args.BluetoothAddress))
                    _addressesBeingInspected.Add(args.BluetoothAddress);
                else
                    alreadyInspectingDevice = true;
            }

            // Skip accessing the device characteristics if it's unsafe because it's also being accessed from another ConsiderDevice at the moment.
            if (Device == null && !alreadyInspectingDevice)
            {
                var device = await BluetoothLEDevice.FromBluetoothAddressAsync(args.BluetoothAddress);

                // Check if the device has the right name
                if (device != null && device.Name.StartsWith(_nameStartsWith))
                {
                    var services = device.GattServices;
                    if (services != null)
                    {
                        // Look for the right service
                        foreach (var service in services)
                        {
                            // Look for the right characteristic
                            foreach (var characteristic in service.GetCharacteristics(_serviceGuid))
                            {
                                // Check if the characteristic has the right properties
                                if (characteristic.CharacteristicProperties.HasFlag(GattCharacteristicProperties.Notify))
                                {
                                    var status = await characteristic.WriteClientCharacteristicConfigurationDescriptorAsync(GattClientCharacteristicConfigurationDescriptorValue.Notify);
                                    if (status == GattCommunicationStatus.Success)
                                    {
                                        device.ConnectionStatusChanged += ConnectionChanged;
                                        characteristic.ValueChanged += ReceivedData;

                                        // Call the OnConnect callback
                                        if (OnConnect != null)
                                            OnConnect(device.Name);

                                        Device = device;
                                        IsInspecting = false;
                                        _advertisementWatcher.Stop();

                                        return;
                                    }
                                }
                            }
                        }

                        foreach (var service in services)
                            service.Dispose();
                    }

                    device.Dispose();
                }

                lock (_addressesBeingInspected)
                {
                    _addressesBeingInspected.Remove(args.BluetoothAddress);
                }
            }
        }

        // Bluetooth LE API watcher, notifying the BluetoothConnection when a new device is found
        private readonly BluetoothLEAdvertisementWatcher _advertisementWatcher = new BluetoothLEAdvertisementWatcher();

        // List if devices currently being inspected
        private readonly IList<ulong> _addressesBeingInspected = new List<ulong>();

        // Start of the Bluetooth device name to look for
        private readonly String _nameStartsWith;

        // Service to look for
        private readonly Guid _serviceGuid;
    }
}