﻿//======================================================================================================================
// NunchuckVJoy project, version 1.0
// Giliam de Carpentier, Copyright (c) 2019.
// See decarpentier.nl/making-a-wireless-joystick for project details.
//======================================================================================================================
// Source code is licensed under the MIT license. 
// See mit.txt for details.
// Use at your own risk!
//======================================================================================================================


using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Threading.Tasks;
using vJoyInterfaceWrap;

namespace NunchuckVJoy
{
    public class VJoystick
    {
        public VJoystick(uint vJoyId)
        {
            _vJoyId = vJoyId;
        }

        public bool Start()
        {
            // Make sure we don't start an already started instance
            if (_vJoy != null)
                return false;

            // Create the vJoy object
            _vJoy = new vJoy();

            // Check if the object is connected to a driver
            if (!_vJoy.vJoyEnabled())
            {
                Console.WriteLine("No vJoy driver was found. Please reinstall vJoy v2.1.8 (or higher).");
                return false;
            }
            else
                Console.WriteLine("Found vJoy driver: {0} {1} by {2}", _vJoy.GetvJoyProductString(), _vJoy.GetvJoySerialNumberString(), _vJoy.GetvJoyManufacturerString());

            // Check if we should be able to actively feed input to vJoy.
            VjdStat status = _vJoy.GetVJDStatus(_vJoyId);
            switch (status)
            {
                case VjdStat.VJD_STAT_OWN:
                    break;
                case VjdStat.VJD_STAT_FREE:
                    break;
                case VjdStat.VJD_STAT_BUSY:
                    Console.WriteLine("vJoy failed: Device {0} already in use", _vJoyId);
                    return false;
                case VjdStat.VJD_STAT_MISS:
                    Console.WriteLine("vJoy failed: Device {0} isn't available/enabled", _vJoyId);
                    return false;
                default:
                    Console.WriteLine("vJoy failed: Device {0} caused unknown error", _vJoyId);
                    return false;
            };

            // Grab the feed to device with the id _vJoyId.
            if (!_vJoy.AcquireVJD(_vJoyId))
            {
                Console.WriteLine("vJoy failed: Device {0} caused unknown error", _vJoyId);
                return false;
            }

            // Reset this device to default values asap
            _vJoy.ResetVJD(_vJoyId);

            // Get the numeric range for the X and Y directions
            _maxAxisXValue = 0;
            if (_vJoy.GetVJDAxisExist(_vJoyId, HID_USAGES.HID_USAGE_X))
                _vJoy.GetVJDAxisMax(_vJoyId, HID_USAGES.HID_USAGE_X, ref _maxAxisXValue);

            _maxAxisYValue = 0;
            if (_vJoy.GetVJDAxisExist(_vJoyId, HID_USAGES.HID_USAGE_Y))
                _vJoy.GetVJDAxisMax(_vJoyId, HID_USAGES.HID_USAGE_Y, ref _maxAxisYValue);

            // Get which buttons are available
            _numButtons = _vJoy.GetVJDButtonNumber(_vJoyId);

            // Do a manual re-center of all axis, because reset vjd doesn't seem to work
            foreach (var usage in (HID_USAGES[])Enum.GetValues(typeof(HID_USAGES)))
            {
                if (_vJoy.GetVJDAxisExist(_vJoyId, usage))
                {
                    long max = 0;
                    _vJoy.GetVJDAxisMax(_vJoyId, usage, ref max);
                    _vJoy.SetAxis((int)max / 2, _vJoyId, usage);
                }
            }

            // Do a manual re-center of discrete hats
            int discCount = _vJoy.GetVJDDiscPovNumber(_vJoyId);
            for (int index = 1; index <= discCount; ++index)
            {
                _vJoy.SetDiscPov(0, _vJoyId, (uint)index);
            }

            // Do a manual re-center of continuous hats
            int contCount = _vJoy.GetVJDContPovNumber(_vJoyId);
            for (int index = 1; index <= contCount; ++index)
            {
                _vJoy.SetContPov(0, _vJoyId, (uint)index);
            }

            // Do a manual release of all buttons
            for (int index = 1; index <= _numButtons; ++index)
            {
                _vJoy.SetBtn(false, _vJoyId, (uint)index);
            }

            Console.WriteLine("Connected to vJoy device {0}", _vJoyId);

            // Success
            return true;
        }

        public void Update(float axisX, float axisY, bool button0, bool button1)
        {
            TrySetAxis(axisX, HID_USAGES.HID_USAGE_X, _maxAxisXValue);
            TrySetAxis(axisY, HID_USAGES.HID_USAGE_Y, _maxAxisYValue);
            TrySetButton(1, button0);
            TrySetButton(2, button1);
        }

        private void TrySetAxis(float normalized, HID_USAGES usage, long max)
        {
            if (usage == HID_USAGES.HID_USAGE_Y)
                normalized = -normalized;

            if (_vJoy != null && max > 0)
            {
                float uniform = 0.5f * normalized + 0.5f;
                int integer = Math.Max(0, Math.Min((int)max, (int)(max * uniform)));
                _vJoy.SetAxis(integer, _vJoyId, usage);
            }
        }

        private void TrySetButton(uint index, bool pressed)
        {
            if (_vJoy != null && index < _numButtons)
                _vJoy.SetBtn(pressed, _vJoyId, index);
        }

        public void Stop()
        {
            _vJoy = null;
        }

        private long _maxAxisXValue;
        private long _maxAxisYValue;
        private int  _numButtons;
        private uint _vJoyId;
        private vJoy _vJoy;
    }
}